from __future__ import annotations

import contextvars
import weakref
from contextlib import contextmanager, suppress
from pathlib import Path
from typing import Generator, Optional, Set, Union

from git import Repo

from octoflow import logging
from octoflow.tracking import (
    Run,
    SQLAlchemyTrackingStore,
    TrackingClient,
)
from octoflow.utils.rsync import rsync

logger = logging.get_logger(__name__)


run_var = contextvars.ContextVar[Optional[Run]]("run", default=None)


class ProjectExperiment:
    def __init__(self, project: Project, expr_name: str) -> None:
        self.get_project = weakref.ref(project)
        self.expr_name = expr_name

    @property
    def project(self) -> Project:
        return self.get_project()

    @contextmanager
    def start_run(
        self,
        force: bool = False,
        description: Optional[str] = None,
    ) -> Generator[Run, None, None]:
        commit_hash = self.project.sync()
        tracking_uri_path = (
            self.project.base_path
            / "experiments"
            / self.expr_name
            / f"{commit_hash}.db"
        )
        if tracking_uri_path.exists():
            if not force:
                msg = f"experiment {self.expr_name} has already been run"
                raise FileExistsError(msg)
            tracking_uri_path.unlink()
        # create the parent directories
        tracking_uri_path.parent.mkdir(parents=True, exist_ok=True)
        tracking_uri = f"sqlite:///{tracking_uri_path}"
        store = SQLAlchemyTrackingStore(tracking_uri)
        client = TrackingClient(store)
        expr = client.get_or_create_experiment(self.expr_name)
        return expr.start_run(commit_hash, description=description)


class ProjectExperimentDict:
    def __init__(self, project: Project) -> None:
        self.get_project = weakref.ref(project)

    @property
    def project(self) -> Project:
        return self.get_project()

    @property
    def experiments_path(self) -> Path:
        exprs_dir = self.project.base_path / "experiments"
        exprs_dir.mkdir(exist_ok=True)
        return exprs_dir

    @property
    def names(self) -> Set[str]:
        return set(
            self.experiments_path.iterdir()
            if self.experiments_path.exists()
            else []
        )

    def __iter__(self):
        yield from self.names

    def __getitem__(self, key: str) -> ProjectExperiment:
        return ProjectExperiment(self.project, key)

    def __contains__(self, key: str) -> bool:
        return key in self.names

    def __len__(self) -> int:
        return len(self.names)

    def __repr__(self) -> str:
        return f"ExperimentDict({self.project.base_path})"


class Project:
    def __init__(self, path: Union[str, Path]) -> None:
        # Initialize the project path
        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)
        # name of the project - different from name of experiment
        self.name = path.name
        self.base_path = path / ".octoflow"
        # Initialize the git repository
        repo = Repo.init(self.base_path / "project.git")
        # commit the initial changes to main if there are no commits
        has_no_commits = True
        with suppress(StopIteration, ValueError):
            next(repo.iter_commits())
            has_no_commits = False
        if has_no_commits:
            repo.index.add("*")
            repo.index.commit("Initialize project")
        self.repo = repo
        self.sync()

    def sync(self, message: Optional[str] = None) -> str:
        # use rsync to copy the project structure
        for cout in rsync(
            self.base_path.parent,
            self.base_path / "project.git",
            exclude=[
                ".git",
                ".gitignore",
                ".octoflow",
            ],
            append_dir=False,
        ):
            logger.info(cout)
        # Commit the changes to the experiment branch
        if "nothing to commit" in self.repo.git.status():
            return self.repo.git.rev_parse("HEAD")
        self.repo.index.add("*")
        if message is None:
            message = "Update project"
        self.repo.index.commit(message)
        return self.repo.git.rev_parse("HEAD")

    @property
    def experiments(self):
        return ProjectExperimentDict(self)



from typing import Any, List, Union

import dill  # noqa: S403
import xxhash

__all__ = [
    "hash",
    "register",
]

dispatch = {}


def register(type_):
    def decorator(func):
        dispatch[type_] = func
        return func

    return decorator


class Hasher:
    """Hasher that accepts python objets as inputs."""

    def __init__(self):
        self.m = xxhash.xxh64()

    @classmethod
    def hash_bytes(cls, value: Union[bytes, List[bytes]]) -> str:
        value = [value] if isinstance(value, bytes) else value
        m = xxhash.xxh64()
        for x in value:
            m.update(x)
        return m.hexdigest()

    @classmethod
    def hash_default(cls, value: Any) -> str:
        return cls.hash_bytes(dill.dumps(value))

    @classmethod
    def hash(cls, value: Any) -> str:
        if type(value) in dispatch:
            return dispatch[type(value)](cls, value)
        else:
            return cls.hash_default(value)

    def update(self, value: Any) -> None:
        header_for_update = f"=={type(value)}=="
        value_for_update = self.hash(value)
        self.m.update(header_for_update.encode("utf-8"))
        self.m.update(value_for_update.encode("utf-8"))

    def hexdigest(self) -> str:
        return self.m.hexdigest()


def hash(obj: Any) -> str:
    h = Hasher()
    h.update(obj)
    return h.hexdigest()



import shutil
import tempfile
from contextlib import suppress
from pathlib import Path
from typing import Optional, Union

from octoflow import config

__all__ = [
    "get_resources_path",
    "get_cache_path",
    "cache",
]


def get_resources_path(path: Optional[str] = None) -> Path:
    if path is None:
        # supress error
        with suppress(AttributeError):
            path = config.resources.path
    return Path(path).expanduser()


def get_cache_path(path: Union[str, Path, None] = None) -> Path:
    if path is None:
        with suppress(AttributeError):
            path = config.resources.cache.path
    if path is not None:
        return Path(path).expanduser()
    try:
        resources_path = get_resources_path()
    except TypeError:
        resources_path = Path(tempfile.gettempdir()) / "octoflow"
    return resources_path / "cache"


class Cache:
    def __init__(self, path: Union[str, Path, None] = None) -> None:
        self._path = get_cache_path(path)

    @property
    def path(self) -> Path:
        return self._path

    def clean(self):
        if self.path.exists():
            shutil.rmtree(self.path)


cache = Cache()



from __future__ import annotations

from typing import Generic, TypeVar

import pyarrow.dataset as ds
from typing_extensions import ParamSpec

__all__ = [
    "PyArrowWrapper",
    "BaseExpression",
    "BaseDataset",
    "BaseDatasetLoader",
]

ArrowType = TypeVar("ArrowType")
P = ParamSpec("P")
R = TypeVar("R")


class PyArrowWrapper(Generic[ArrowType]):
    def __init__(self, wrapped: ArrowType) -> None:
        self._wrapped = wrapped

    def to_pyarrow(self) -> ArrowType:
        return self._wrapped


BaseExpression = PyArrowWrapper[ds.Expression]

BaseDataset = PyArrowWrapper[ds.Dataset]


class BaseDatasetLoader(Generic[P, R]): ...



import re

ESCAPED_CHARS = r"\\`*_{}[]()#+\-.!"
ESCAPED_CHARS_RE = re.compile(r"([%s])" % re.escape(ESCAPED_CHARS))
UNESCAPED_CHARS_RE = re.compile(r"\\([%s])" % re.escape(ESCAPED_CHARS))


def escape(text: str) -> str:
    """Escape text."""
    return ESCAPED_CHARS_RE.sub(r"\\\1", text)


def unescape(text: str) -> str:
    """Unescape text."""
    return UNESCAPED_CHARS_RE.sub(r"\1", text)



import itertools
from typing import Any, Dict, Generator, Optional, Tuple, TypeVar

import pyarrow as pa
from typing_extensions import Self

from octoflow.data.metadata import unify_metadata
from octoflow.data.types import infer_type, unify_types
from octoflow.exceptions import ValidationError

__all__ = [
    "unify_schemas",
    "infer_schema",
]

T = TypeVar("T")


class SchemaBuilder:
    def __init__(self):
        self._fields = {}
        # Keys and values must be coercible to bytes.
        self.metadata = None

    def has_field(self, name: str) -> bool:
        return name in self._fields

    def get_field(self, name: str) -> pa.Field:
        return self._fields[name]

    def update_field(
        self,
        name: str,
        dtype: pa.DataType,
        nullable: bool = True,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        self._fields[name] = pa.field(name, dtype, nullable, metadata)

    def remove_field(self, name: str):
        del self._fields[name]

    def build(self) -> pa.Schema:
        fields = list(self._fields.values())
        return pa.schema(fields, metadata=self.metadata)


def unify_schemas(this: pa.Schema, other: Optional[pa.Schema]) -> pa.Schema:
    if other is None:
        return this
    builder = SchemaBuilder()
    this_schema_fields = set()
    for this_field in this:
        this_field_name = this_field.name
        builder.update_field(
            this_field.name,
            this_field.type,
            this_field.nullable,
            this_field.metadata,
        )
        this_schema_fields.add(this_field_name)
    for other_field in other:
        other_field_name = other_field.name
        this_schema_fields.discard(other_field_name)
        if builder.has_field(other_field_name):
            this_field = builder.get_field(other_field_name)
            try:
                promoted_type = unify_types(this_field.type, other_field.type)
            except ValueError as ex:
                msg = (
                    f"cannot unify types '{this_field.type}' and "
                    f"'{other_field.type}' of field '{other_field_name}'"
                )
                raise ValueError(msg) from ex
            nullable = (
                this_field.nullable
                or this_field.type.equals(pa.null())
                or other_field.nullable
                or other_field.type.equals(pa.null())
            )
            metadata = unify_metadata(this_field, other_field)
            builder.update_field(
                other_field_name,
                promoted_type,
                nullable,
                metadata,
            )
        else:
            builder.update_field(
                other_field_name,
                other_field.type,
                True,  # nullable because this field is not in the existing
                other_field.metadata,
            )
    for this_field_name in this_schema_fields:
        this_field = builder.get_field(this_field_name)
        builder.update_field(
            this_field_name,
            this_field.type,
            True,  # nullable because this field is not in the other schema
            this_field.metadata,
        )
    builder.metadata = unify_metadata(this, other)
    return builder.build()


def infer_schema(data: Dict[str, Any]) -> Self:
    schema_builder = SchemaBuilder()
    for key, value in data.items():
        pa_type = infer_type(value)
        schema_builder.update_field(key, pa_type)
    return schema_builder.build()


def validate(schema: pa.Schema, data: dict) -> bool:
    """
    Validates a dictionary against a PyArrow schema.

    Parameters
    ----------
    schema : pyarrow.Schema
        The PyArrow schema to validate against.
    data : dict
        The dictionary to validate.

    Raises
    ------
    ValidationError
        If the dictionary does not match the schema.

    Examples
    --------
    >>> schema = pa.schema([pa.field('id', pa.int64()), pa.field('name', pa.string())])
    >>> valid_dict = {'id': 1, 'name': 'Alice'}
    >>> validate(schema, valid_dict)
    >>> invalid_dict = {'id': '1', 'name': 'Alice'}
    >>> validate(schema, invalid_dict)
    Traceback (most recent call last):
    ...
    ValidationError: ...
    """  # noqa: E501
    try:
        pa.RecordBatch.from_pylist([data], schema)
    except pa.lib.ArrowInvalid as e:
        raise ValidationError(str(e).lower()) from e


def get_schema(data: T) -> Tuple[T, pa.Schema]:
    try:
        return data, data.schema
    except AttributeError as e:
        if not isinstance(data, Generator):
            msg = (
                "expected data to be of type 'Generator', "
                f"got '{data.__class__.__name__}'"
            )
            raise TypeError(msg) from e
        data, data_ = itertools.tee(data)
        return data, next(data_).schema



from __future__ import annotations

from dataclasses import MISSING
from operator import itemgetter
from typing import (
    Any,
    Dict,
    Generator,
    List,
    Mapping,
    Optional,
    Sequence,
    Union,
    overload,
)

import orjson
import pyarrow as pa
from typing_extensions import Self

from octoflow.data.field import Field
from octoflow.data.schema import infer_schema, unify_schemas
from octoflow.data.types import UNDEFINED, infer_type, unify_types
from octoflow.utils.escape import escape, unescape

__all__ = [
    "Document",
    "DocumentBatch",
]


class Document(Mapping[str, Any]):
    """Document class."""

    id: Any
    _id_type: pa.DataType
    _data: Mapping[str, Any]
    _schema: pa.Schema

    def __init__(
        self,
        data: Mapping[str, Any],
        /,
        id: Union[itemgetter, Field, Any] = MISSING,
        *,
        schema: Union[pa.Schema, None] = None,
    ):
        """Create a new document.

        Parameters
        ----------
        data : dict
            The data of the document.
        id : Union[itemgetter, Any], optional
            The id of the document, by default MISSING.
        schema : dict, optional
            The schema of the document, by default None.
        """
        # Data (must be json serializable with orjson)
        if not isinstance(data, dict):
            msg = "data must be a dictionary"
            raise ValueError(msg)
        data = orjson.loads(orjson.dumps(data))
        # Id
        if id is MISSING:
            id = Field("id")
        elif isinstance(id, itemgetter):
            id = Field(id)
        id_type = None
        if isinstance(id, Field):
            id_type = id.type
            id = id(data)
        if id_type is None:
            id_type = infer_type(id)
        # Schema
        if not isinstance(schema, pa.Schema):
            schema = infer_schema(data)
        self.id = id
        self._id_type = id_type
        self._data = data
        self._schema = schema

    @property
    def schema(self) -> pa.Schema:
        """Return the schema of the document.

        Returns
        -------
        Schema
            The schema of the document.
        """
        return self._schema

    def __getitem__(self, key: str):
        """Get the value of a field.

        Parameters
        ----------
        key : str
            The field to be accessed.

        Returns
        -------
        Any
            The value of the field.
        """
        return self._data[key]

    def __iter__(self):
        """Iterate over the fields of the document.

        Returns
        -------
        iter
            An iterator over the fields of the document.
        """
        return iter(self._data)

    def __len__(self):
        """Return the number of fields in the document.

        Returns
        -------
        int
            The number of fields in the document.
        """
        return len(self._data)

    def __repr__(self):
        """Return a string representation of the document.

        Returns
        -------
        str
            The string representation of the document.
        """
        return f"Document({self._data})"


class DocumentBatch(Sequence[Document]):
    """DocumentBatch class -- list version of a table."""

    _ids: List[Any]
    _data: Dict[str, List[Any]]
    _schema: Union[pa.Schema, None]
    _id_index: Dict[int, int]
    _id_type: pa.DataType
    _length: int

    def __init__(self, docs: Optional[Sequence[Document]] = None):
        """Create a new document index."""
        self._length = 0
        self._ids = []
        self._data = {}
        self._id_index = {}
        self._schema = None
        self._id_type = UNDEFINED
        if docs is not None:
            self.extend(docs)

    @property
    def schema(self) -> pa.Schema:
        """Return the schema of the document list.

        Returns
        -------
        Schema
            The schema of the document list.
        """
        return self._schema

    @overload
    def __getitem__(self, item: slice) -> DocumentBatch: ...

    @overload
    def __getitem__(self, item: int) -> Document: ...

    def __getitem__(
        self, item: Union[int, slice]
    ) -> Union[Document, DocumentBatch]:
        """Get a document by index.

        Parameters
        ----------
        item : int
            The index of the document to be accessed.

        Returns
        -------
        Document
            The document at the specified index.
        """
        if isinstance(item, slice):
            dlist = self.__class__()
            # filter the data
            dlist._data = {k: v[item] for k, v in self._data.items()}
            # filter the ids
            dlist._ids = self._ids[item]
            dlist._id_index = {id: i for i, id in enumerate(dlist._ids)}
            # copy the schema and id type
            dlist._schema = self._schema
            dlist._id_type = self._id_type
            return dlist
        return Document(
            {k: v[item] for k, v in self._data.items()},
            id=self._ids[item],
            schema=self._schema,
        )

    def append(self, doc: Document):
        """Add a document to the index.

        Parameters
        ----------
        doc : Document
            The document to be added.

        Raises
        ------
        ValueError
            If the document already exists in the index.
        """
        # validate the document
        if doc.id in self._id_index:
            msg = f"document with id {doc.id} already exists"
            raise ValueError(msg)
        unified_schema = unify_schemas(doc.schema, self._schema)
        unified_id_type = unify_types(self._id_type, doc._id_type)
        # update the list
        self._schema = unified_schema
        self._id_type = unified_id_type
        for col in self._schema.names:
            self._data.setdefault(
                col, [None for _ in range(self._length)]
            ).append(doc.get(col, None))
        self._ids.append(doc.id)
        self._id_index[doc.id] = self._length
        self._length += 1

    def extend(self, docs: Sequence[Document]):
        """Add multiple documents to the index.

        Parameters
        ----------
        docs : Sequence[Document]
            The documents to be added.

        Raises
        ------
        ValueError
            If any document already exists in the index.
        """
        for doc in docs:
            self.append(doc)

    def __len__(self) -> int:
        """Return the number of documents in the index.

        Returns
        -------
        int
            The number of documents in the index.
        """
        return self._length

    def __iter__(self) -> Generator[Document, None, None]:
        """Iterate over the documents in the index.

        Returns
        -------
        iter
            An iterator over the documents in the index.
        """
        for i in range(self._length):
            yield self[i]

    def __contains__(self, item: Document) -> bool:
        """Check if a document is in the index.

        Parameters
        ----------
        item : Document
            The document to be checked.

        Returns
        -------
        bool
            True if the document is in the index, False otherwise.
        """
        return item.id in self._id_index

    def __repr__(self) -> str:
        """Return a string representation of the document list.

        Returns
        -------
        str
            The string representation of the document list.
        """
        if self._length > 3:  # noqa: PLR2004
            sample = self._data[:2] + ["...", self._data[-1]]
        else:
            sample = self._data
        return f"DocumentList[{self._length}]({sample})"

    def to_arrow(self) -> pa.RecordBatch:
        """Return the document list as a table.

        Returns
        -------
        pa.RecordBatch
            The document list as a RecordBatch.
        """
        id_field = pa.field("_id", self._id_type)
        fields = [id_field]
        arrays = [pa.array(self._ids, type=self._id_type)]
        for col in self._schema.names:
            f = self._schema.field(col)
            fields.append(f.with_name(escape(f.name)))
            arrays.append(pa.array(self._data[col], type=f.type))
        return pa.record_batch(arrays, schema=pa.schema(fields))

    @classmethod
    def from_arrow(cls, batch: Union[pa.RecordBatch, pa.Table]) -> Self:
        """Create a document batch from a pyarrow.RecordBatch or pyarrow.Table.

        Parameters
        ----------
        batch : pa.RecordBatch | pa.Table
            The table to be converted to a document batch.

        Returns
        -------
        DocumentBatch
            The document list.
        """
        this = cls.__new__(cls)
        id_field_index = batch.schema.get_field_index("_id")
        if id_field_index < 0:
            msg = "missing '_id' field"
            raise ValueError(msg)
        id_col = batch.column(id_field_index)
        this._ids = id_col.to_pylist()
        columns = [(f, unescape(f)) for f in batch.schema.names if f != "_id"]
        this._data = {
            key: batch.column(col).to_pylist() for col, key in columns
        }
        this._schema = pa.schema([
            batch.schema.field(col).with_name(key) for col, key in columns
        ])
        this._id_index = {id: i for i, id in enumerate(this._ids)}
        this._id_type = id_col.type
        this._length = id_col.length()
        return this



from __future__ import annotations

import os
import tempfile
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any, Generator, Iterable, List, Optional, Union

import pandas as pd
import pyarrow as pa
import pyarrow.dataset as ds

from octoflow import logging
from octoflow.data.schema import get_schema
from octoflow.utils import hashutils, resources

logger = logging.get_logger(__name__)


def generate_unique_path(
    reference: Any, cache_dir: Optional[Union[str, Path]] = None
) -> Path:
    # create a temporary directory in system temp directory
    if cache_dir is None:
        cache_dir = resources.get_cache_path() / "datasets"
    if cache_dir is not None and not cache_dir.exists():
        try:
            cache_dir.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            msg = f"failed to create cache directory at '{cache_dir}'"
            raise OSError(msg) from e
    try:
        fingerprint = hashutils.hash(reference)
        path = cache_dir / f"data-{fingerprint}"
    except Exception as e:
        # an error occurred while hashing data
        msg = "failed to create unique path, using temporary directory"
        msg = e.args[0] if e.args else msg
        logger.warning(msg)
        path = Path(tempfile.mkdtemp(dir=cache_dir))
    return path


def get_data_path(path: Union[Path, str]) -> Path:
    if isinstance(path, str):
        path = Path(path)
    if path.is_dir() or not path.exists():
        path.mkdir(
            parents=True,
            exist_ok=True,
        )
        data_path = path / "data"
    else:
        msg = f"expected path to be directory, got '{path}'"
        raise ValueError(msg)
    return data_path


def record_batch(data: Any) -> Union[pa.RecordBatch, pa.RecordBatchReader]:
    # if generator, return a generator
    if isinstance(data, pa.RecordBatch):
        return data
    if isinstance(data, Generator):
        data = (record_batch(d) for d in data)
        data, schema = get_schema(data)
        return pa.RecordBatchReader.from_batches(schema, data)
    if isinstance(data, pa.Table):
        # zero-copy
        return pa.RecordBatchReader.from_batches(
            data.schema, (d for d in data.to_batches())
        )
    if isinstance(data, pd.DataFrame):
        return pa.RecordBatch.from_pandas(data)
    if isinstance(data, Mapping):
        return pa.RecordBatch.from_pydict(data)
    if isinstance(data, Sequence) and not isinstance(data, str):
        return pa.RecordBatch.from_pylist(list(data))
    dtype = data.__class__.__name__
    msg = (
        f"expected data to be of type 'pa.RecordBatch', 'pa.Table', "
        f"'pd.DataFrame', Mapping, or Sequence, got '{dtype}'"
    )
    raise ValueError(msg)


def write_dataset(
    path: Union[str, Path],
    data: Union[
        ds.Dataset,
        pa.Table,
        pa.RecordBatch,
        # Iterable[pa.Table],
        Iterable[pa.RecordBatch],  # pa.RecordBatchReader.from_batches
        pa.RecordBatchReader,
        pd.DataFrame,  # from_pandas
        Mapping[str, List[Any]],  # from_pydict
        Sequence[Mapping[str, Any]],  # from_pylist
    ],
    schema: pa.Schema = None,
    format: Optional[str] = None,
) -> bool:
    if isinstance(path, str):
        path = Path(path)
    if not path.exists():
        path.mkdir(
            parents=True,
            exist_ok=True,
        )
    out_data_path = get_data_path(path)
    if out_data_path.exists():
        return False
    # first write to temporary directory
    temp_path = Path(
        tempfile.mkdtemp(
            prefix=".temp-",
            dir=path,
        )
    )
    # it might take some time to write the data
    # within that time, another process might try
    # to read the data or write the data so we
    # write to a temporary directory first
    # and then move the data to the desired
    # directory
    ds.write_dataset(
        data if isinstance(data, ds.Dataset) else record_batch(data),
        temp_path,
        schema=schema,
        format=format,
    )
    try:
        os.replace(temp_path, out_data_path)
    except OSError:
        return False
    return True


def read_dataset(
    path: Union[str, Path],
    format: str,
) -> ds.dataset:
    data_path = get_data_path(path)
    return ds.dataset(
        data_path,
        format=format,
    )



from __future__ import annotations

from dataclasses import Field, dataclass, field
from typing import Any, Optional

from sqlalchemy import Table
from sqlalchemy.orm import registry as registry_cls
from typing_extensions import dataclass_transform, get_type_hints

__all__ = [
    "ModelBase",
    "ModelMeta",
]


@dataclass_transform(field_specifiers=(Field, field))
class ModelMeta(type):
    def __new__(cls, name, bases, attrs, **kwargs):
        cls = super().__new__(cls, name, bases, attrs)
        table: Optional[Table] = kwargs.get(
            "table", getattr(cls, "__table__", None)
        )
        if table is not None and not hasattr(cls, "__table__"):
            cls.__table__ = table
        cls = dataclass(cls)
        registry: registry_cls = kwargs.get("registry", None)
        if table is not None and registry is not None:
            return registry.mapped(cls)
        return cls

    def update_forward_refs(cls, **kwargs: Any) -> None:
        # update forward references in all subclasses
        for subclass in cls.__subclasses__():
            subclass.update_forward_refs(**kwargs)
            # update forward references in type hints
        for name, value in get_type_hints(cls, localns=kwargs).items():
            cls.__annotations__[name] = value


class ModelBase(metaclass=ModelMeta):
    def __post_init__(self):
        pass



from __future__ import annotations

from operator import itemgetter
from typing import TYPE_CHECKING, Any, Mapping, Union

import pyarrow as pa
from typing_extensions import Self

__all__ = [
    "Field",
]


if TYPE_CHECKING:
    pass


class Field:
    def __new__(
        cls, field: Union[itemgetter, str, Self], /, *args, **kwargs
    ) -> Self:
        """Create a new field getter.

        Parameters
        ----------
        field : Union[itemgetter, str, FieldGetter]
            The field to be accessed.
        type : Union[pa.DataType, None], optional
            The type of the field, by default None.
        preprocessor : Union[callable, None], optional
            A function to preprocess the field, by default None.

        Returns
        -------
        FieldGetter
            The field getter.
        """
        if isinstance(field, Field):
            return field
        return super().__new__(cls)

    def __init__(
        self,
        field: Union[itemgetter, str],
        /,
        type: Union[pa.DataType, None] = None,
        preprocessor: Union[callable, None] = None,
    ):
        """Create a new field getter.

        Parameters
        ----------
        field : Union[itemgetter, str, FieldGetter]
            The field to be accessed.
        type : Union[pa.DataType, None], optional
            The type of the field, by default None.
        preprocessor : Union[callable, None], optional
            A function to preprocess the field, by default None.
        """
        if not isinstance(field, itemgetter):
            field = itemgetter(field)
        self.getter = field
        self.type = type
        self.preprocessor = preprocessor

    def __call__(self, data: Mapping[str, Any]) -> Any:
        """Get the value of the field.

        Parameters
        ----------
        data : dict
            The data to be accessed.
        """
        if self.preprocessor is not None:
            return self.preprocessor(self.getter(data))
        return self.getter(data)


def field(
    field: Union[str, itemgetter, Field],
    /,
    type: Union[pa.DataType, None] = None,
    preprocessor: Union[callable, None] = None,
) -> Field:
    """Create a new field getter."""
    return Field(
        field,
        type=type,
        preprocessor=preprocessor,
    )



from __future__ import annotations

import contextlib
import functools
import shutil
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
    overload,
)

import numpy as np
import pandas as pd
import pyarrow as pa
from numpy.typing import ArrayLike
from pandas import DataFrame
from tqdm import auto as tqdm

from octoflow import logging
from octoflow.data.base import BaseDataset
from octoflow.data.constants import DEFAULT_BATCH_SIZE, DEFAULT_FORMAT
from octoflow.data.expression import Expression
from octoflow.data.loaders import DatasetLoader, loaders
from octoflow.data.utils import (
    generate_unique_path,
    read_dataset,
    record_batch,
    write_dataset,
)
from octoflow.utils import hashutils

logger = logging.get_logger(__name__)

SourceType = Union[
    str,
    List[str],
    Union[Path, List[Path]],
    "Dataset",
    List["Dataset"],
]


def _map_func_wrapper(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        output = func(*args, **kwargs)
        if isinstance(output, pd.Series):
            return output
        return pd.Series(output)

    return wrapped


class Dataset(BaseDataset):  # noqa: PLR0904
    @overload
    def __init__(
        self,
        data: Union[List[dict], Dict[str, list], DataFrame] = None,
        format: str = DEFAULT_FORMAT,
    ): ...

    @overload
    def __init__(
        self,
        data: Union[List[dict], Dict[str, list], DataFrame],
        format: str = DEFAULT_FORMAT,
        *,
        path: Optional[Union[str, Path]] = None,
        cache_dir: Optional[Union[str, Path]] = None,
    ): ...

    @overload
    def __init__(
        self,
        loader: DatasetLoader,
        format: str = DEFAULT_FORMAT,
        *,
        path: Optional[Union[str, Path]] = None,
        cache_dir: Optional[Union[str, Path]] = None,
        loader_args: Optional[Tuple[Any, ...]] = None,
        loader_kwargs: Optional[Dict[str, Any]] = None,
    ): ...

    def __init__(
        self,
        data_or_loader: Union[
            List[dict], Dict[str, list], DataFrame, DatasetLoader
        ] = None,
        format: str = DEFAULT_FORMAT,
        *,
        schema: Optional[pa.Schema] = None,
        path: Optional[Union[str, Path]] = None,
        cache_dir: Optional[Union[str, Path]] = None,
        loader_args: Optional[Tuple[Any, ...]] = None,
        loader_kwargs: Optional[Dict[str, Any]] = None,
    ):
        """
        Create a new dataset.

        Parameters
        ----------
        data_or_loader : list of dict, dict of list, DataFrame, BaseDatasetLoader
            The data to load into the dataset.
        format : str
            The format of the dataset.
        path : str, Path, None
            The path to the dataset.
        cache_dir : str, Path, None
            The directory to use for caching.
        loader_args : tuple, None
            The arguments to pass to the loader function if provided
                as the first argument.
        loader_kwargs : dict, None
            The keyword arguments to pass to the loader function if provided
                as the first argument.
        """  # noqa: E501
        if isinstance(data_or_loader, str):
            data_or_loader = loaders[data_or_loader]
        if isinstance(data_or_loader, DatasetLoader):
            if loader_args is None:
                loader_args = ()
            if loader_kwargs is None:
                loader_kwargs = {}
            data_or_loader = data_or_loader.partial(
                *loader_args, **loader_kwargs
            )
            data = data_or_loader()
        else:
            data = data_or_loader
        if path is None:
            path = generate_unique_path(data_or_loader, cache_dir=cache_dir)
        if format is None:
            format = DEFAULT_FORMAT
        created = write_dataset(path, data, schema=schema, format=format)
        if not created:
            msg = (
                f"existing dataset found at '{path}', loading existing file(s)"
            )
            logger.warning(msg)
        dataset = read_dataset(path, format=format)
        super().__init__(dataset)
        self._path = path
        self._format = format

    @property
    def path(self) -> Path:
        """
        The path to the dataset.

        Returns
        -------
        Path
            The path to the dataset.
        """
        return self._path

    @property
    def format(self) -> str:
        """
        The format of the dataset.

        Returns
        -------
        str
            The format of the dataset.
        """
        return self._format

    @property
    def _wrapped_format_default_extname(self) -> str:
        return self._wrapped.format.default_extname

    def count_rows(self) -> int:
        """
        Count the number of rows in the dataset.

        Returns
        -------
        int
            The number of rows in the dataset.
        """
        return self._wrapped.count_rows()

    def head(
        self,
        num_rows: int = 5,
        columns: Union[str, List[str], None] = None,
        filter: Expression = None,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> DataFrame:
        """
        Get the first rows of the dataset as a pandas DataFrame.

        Parameters
        ----------
        num_rows : int
            The number of rows to get.
        columns : str, list of str, None
            Names of columns to get. If None, all columns are returned.
        filter : Expression
            The filter expression.
        batch_size : int
            Number of rows to get at a time.

        Returns
        -------
        DataFrame
            A pandas DataFrame containing the first rows of the dataset.
        """
        filter = filter.to_pyarrow() if filter else None
        if isinstance(columns, str):
            columns = [columns]
        table: pa.Table = self._wrapped.head(
            num_rows=num_rows,
            columns=columns,
            filter=filter,
            batch_size=batch_size,
        )
        return table.to_pandas()

    @overload
    def __getitem__(self, indices: int) -> Dict[str, Any]: ...

    @overload
    def __getitem__(
        self, indices: Union[slice, List[int], ArrayLike]
    ) -> pa.Table: ...

    def __getitem__(
        self, indices: Union[int, slice, List[int], ArrayLike]
    ) -> Union[dict, pa.Table]:
        """Get rows from the dataset."""
        return self.take(indices=indices)

    @overload
    def take(
        self,
        *,
        indices: Optional[int] = None,
        columns: Union[str, List[str], None] = None,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> Dict[str, Any]: ...

    @overload
    def take(
        self,
        *,
        indices: Union[slice, List[int], ArrayLike] = None,
        columns: Union[str, List[str], None] = None,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> DataFrame: ...

    def take(
        self,
        *,
        indices: Union[int, slice, List[int], ArrayLike] = None,
        columns: Union[str, List[str], None] = None,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> Union[dict, pa.Table]:
        """
        Take rows from the dataset.

        Parameters
        ----------
        indices : int, slice, list of int, array-like
            Indices of rows to take.
        columns : str, list of str, None
            Names of columns to take. If None, all columns are taken.
        batch_size : int
            Number of rows to take at a time.

        Returns
        -------
        Document, pyarrow.Table
            The taken rows or row.
        """
        return_batch = True
        if indices is None:
            indices = []
        elif isinstance(indices, int):
            return_batch = False
            indices = [indices]
        elif isinstance(indices, slice):
            step_size = indices.step
            if step_size is None:
                # default step size is 1
                step_size = 1
            elif not isinstance(step_size, int):
                msg = "expected indices.step to be int, got '{indices.step}'"
                raise ValueError(msg)
            elif step_size < 1:
                msg = (
                    "expected indices.step to be greater than 0, "
                    f"got '{step_size}'"
                )
                raise ValueError(msg)
            start = indices.start or 0
            stop = indices.stop or self.count_rows()
            indices = np.arange(start, stop, step_size)
        if isinstance(columns, str):
            columns = [columns]
        batch = self._wrapped.take(
            indices=indices,  # array or array-like
            columns=columns,  # list of str or None
            batch_size=batch_size,  # int
        )
        if return_batch:
            return batch
        return batch.to_pylist()[0]

    def map(
        self,
        func: Any,
        batch_size: int = DEFAULT_BATCH_SIZE,
        batched: bool = False,
        verbose: Union[bool, int] = 1,
    ) -> Dataset:
        """
        Map a function over the dataset.

        Parameters
        ----------
        func : Any
            The function to map over the dataset.
        batch_size : int
            Number of rows to map at a time.
        batched : bool
            Whether the function is batched.
        verbose : bool | int
            Whether to show a progress bar.

        Returns
        -------
        Dataset
            A new dataset containing the mapped rows.
        """
        fingerprint = hashutils.hash(func)
        path = self.path / f"map-{fingerprint}"
        if path.exists():
            logger.warning(
                f"existing dataset found at '{path}', loading"
                " existing file(s)"
            )
        else:
            num_batches = ((self.count_rows() - 1) // batch_size) + 1
            batch_iter = (
                record_batch(
                    func(batch)
                    if batched
                    else batch.to_pandas().apply(
                        _map_func_wrapper(func),
                        axis=1,
                    )
                )
                for batch in tqdm.tqdm(
                    self._wrapped.to_batches(batch_size=batch_size),
                    total=num_batches,
                    desc="Mapping",
                )
            )
            _ = write_dataset(
                path,
                batch_iter if verbose else batch_iter,
                format=self.format,
            )
        return type(self).load_dataset(path, format=self.format)

    def filter(self, expression: Expression = None) -> Dataset:
        """
        Filter the dataset.

        Parameters
        ----------
        expression : Expression
            The filter expression.

        Returns
        -------
        Dataset
            A new dataset containing only the rows that match the filter expression.
        """  # noqa: E501
        if expression is None:
            return self
        pyarrow_expression = expression.to_pyarrow()
        fingerprint = hashutils.hash(pyarrow_expression)
        path = self.path / f"filter-{fingerprint}"
        if path.exists():
            logger.warning(
                f"existing dataset found at '{path}', loading"
                " existing file(s)"
            )
        else:
            dataset = self._wrapped.filter(pyarrow_expression)
            write_dataset(
                path,
                dataset,
                schema=dataset.schema,
                format=self.format,
            )
        return self.load_dataset(path, format=self.format)

    def cleanup(self):
        """
        Delete the directory containing the dataset.

        Returns
        -------
        None
        """
        if not self.path.exists():
            return
        shutil.rmtree(self.path)

    def __len__(self) -> int:
        """
        Get the number of rows in the dataset.

        Returns
        -------
        int
            The number of rows in the dataset.
        """
        return self.count_rows()

    def __enter__(self) -> Dataset:
        """
        Context manager entry point.

        Returns
        -------
        Dataset
            The dataset.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Context manager exit point.

        This will delete the resources allocated for this dataset.

        Parameters
        ----------
        exc_type : Exception
            The exception type.
        exc_value : Exception
            The exception value.
        traceback : Traceback
            The traceback.

        Returns
        -------
        None
        """
        with contextlib.suppress(Exception):
            self.cleanup()

    @classmethod
    def load_dataset(
        cls,
        path: Union[Path, str],
        format: str = DEFAULT_FORMAT,
    ) -> Dataset:
        """
        Load an existing dataset.

        Parameters
        ----------
        path : str, Path
            The path to the dataset.
        format : str
            The format of the dataset.

        Returns
        -------
        Dataset
            The loaded dataset.
        """
        inst = cls.__new__(cls)
        inst._wrapped = read_dataset(path, format=format)
        inst._path = path
        inst._format = format
        return inst
